# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntFlag

import gitlab
from git import Repo

from git_system_follower.logger import logger
from git_system_follower.typings.repository import RepositoryInfo


__all__ = ['checkout_to_new_branch', 'push_installed_packages']


class PushFlags(IntFlag):
    NEW_TAG = 1
    NEW_HEAD = 2
    NO_MATCH = 4
    REJECTED = 8
    REMOTE_REJECTED = 16
    REMOTE_FAILURE = 32
    DELETED = 64
    FORCED_UPDATE = 128
    FAST_FORWARD = 256
    UP_TO_DATE = 512
    ERROR = 1024

def checkout_to_new_branch(repo: Repo, base_branch: str) -> str:
    new_branch = f'{base_branch}.temp-manage-packages'
    repo.git.checkout(base_branch)
    repo.remotes.origin.pull(base_branch)

    for branch in repo.heads:
        if branch.name == new_branch:
            repo.delete_head(new_branch, force=True)
            logger.debug(f'Local branch {new_branch} deleted')
            break

    repo.git.checkout('-b', new_branch)
    logger.success(f'Created new {new_branch} local branch (local repo: {repo.git.working_dir})')
    return new_branch


def push_installed_packages(repo: RepositoryInfo, msg: str, *, name: str, email: str) -> None:
    """ Push changes to remote repository

    :param repo: repo information
    :param msg: commit message
    :param name: user name for commit changes
    :param email: user email for commit changes
    """
    try:
        repo.gitlab.branches.delete(repo.git.active_branch.name)
    except gitlab.exceptions.GitlabDeleteError:
        # if temp branch does not exist
        pass
    repo.git.git.add(A=True)
    repo.git.config_writer().set_value('user', 'name', name).release()
    repo.git.config_writer().set_value('user', 'email', email).release()
    repo.git.index.commit(msg)
    result = interpret_push_info(repo.git.remotes.origin.push(repo.git.active_branch.name)[0])
    if not result['flags']:
        logger.critical(f"[{result['reason']}] Push to remote repository failed with reason: {result['summary']}")
        raise SystemExit

def interpret_push_info(push_info) -> dict:
    """ Interpret GitPython push_info flags and return structured result.
    For more details see https://gitpython.readthedocs.io/en/stable/reference.html#git.remote.PushInfo

    :param push_info: A GitPython PushInfo object
    :return: dict with status flags, reason, and summary
    """
    flags = PushFlags(push_info.flags)
    summary = push_info.summary

    status_checks = [
        (PushFlags.ERROR not in flags, True, "SUCCESS"),
        (PushFlags.UP_TO_DATE in flags, True, "UP_TO_DATE"),
        (PushFlags.NEW_HEAD in flags, True, "NEW_HEAD"),
        (PushFlags.NEW_TAG in flags, True, "NEW_TAG"),
        (PushFlags.FAST_FORWARD in flags, True, "FAST_FORWARD"),
        (PushFlags.FORCED_UPDATE in flags, True, "FORCED_UPDATE"),
        (PushFlags.DELETED in flags, False, "DELETED"),
        (PushFlags.REJECTED in flags, False, "REJECTED"),
        (PushFlags.NO_MATCH in flags, False, "NO_MATCH"),
        (PushFlags.REMOTE_FAILURE in flags, False, "REMOTE_FAILURE"),
        (PushFlags.REMOTE_REJECTED in flags, False, "REMOTE_REJECTED"),
        (PushFlags.ERROR in flags, False, "ERROR"),
    ]

    for condition, is_success, reason in status_checks:
        if condition:
            return {
                "flags": is_success,
                "reason": reason,
                "summary": summary
            }

    return {
        "flags": False,
        "reason": "OTHERS",
        "summary": summary
    }



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with output info related utils """
from typing import Iterable, Callable

from outlify.panel import ParamsPanel
from outlify.style import AnsiCodes, Colors, Styles

from git_system_follower.logger import logger
from git_system_follower.typings.package import PackageLocalData


__all__ = ['BrandedColors', 'banner', 'display_params', 'print_dependency_tree_one_level']


WIDTH = 100


class BrandedAnsiCodes(AnsiCodes):
    orange = (38, 2, 244, 81, 30)  # color: #F4511E


BrandedColors = BrandedAnsiCodes()
COMMON_SETTINGS = {
    'width': 100,
    'title_style': [Colors.white],
    'title_align': 'left',
    'subtitle_style': [Colors.gray],
    'subtitle_align': 'right',
    'border': '╭╮╰╯═',
    'border_style': [Colors.gray],
}


def banner(version: str, *, output_func: Callable = print):
    content = f"""
    {BrandedColors.orange}.-,{Colors.reset}
 {BrandedColors.orange}.^.: :.^.{Colors.reset}   ┏┓╻┳ ┏┓╻╻┏┓┳┏┓┏┳┓ ┏┓┏┓╻ ╻ ┏┓┏ ┓┏┓┳┓
{BrandedColors.orange},-' .-. '-,{Colors.reset}  ┃┓┃┃ ┗┓┗┃┗┓┃┣ ┃┃┃ ┣ ┃┃┃ ┃ ┃┃┃┃┃┣ ┣┛
{BrandedColors.orange}'-. '-' .-'{Colors.reset}  ┗┛╹╹ ┗┛┗┛┗┛╹┗┛╹ ╹ ╹ ┗┛┗┛┗┛┗┛┗┻┛┗┛┛┗
 {BrandedColors.orange}'.`; ;`.'{Colors.reset}   {Colors.white}{Styles.bold}git-system-follower{Styles.reset} v{version}
    {BrandedColors.orange}`-`{Colors.reset}"""
    output_func(content)


def display_params(data: dict[str, dict]) -> None:
    """ Display parameters in a compact format

    :param data: data to display where key - section name, value - parameters
    """
    for i, (name, params) in enumerate(data.items(), 1):
        title, subtitle = f'{i}. {name} parameters', f'total: {len(params)}'
        logger.info(ParamsPanel(params, title=title, subtitle=subtitle, **COMMON_SETTINGS))


def print_dependency_tree_one_level(
        packages: Iterable[PackageLocalData], title='', *,
        key: Callable, output_func: Callable = print
) -> None:
    """ Print dependency tree

    :param packages: packages which need to print
    :param title: title of tree
    :param key: function for filtering the information from the list
    :param output_func: output function
    """
    content = f'{title}:\n'
    for i, package in enumerate(packages, 1):
        content += f'{i}. {key(package)}\n'
        prefix = ' ' * len(str(i)) + '  '  # spaces before connector to level the tree
        for j, dependency in enumerate(package['dependencies']):
            connector = '└── ' if j == len(package['dependencies']) - 1 else '├── '
            content += f'{prefix}{connector}{dependency}\n'

    if content[-1] == '\n':
        content = content[:-1]
    output_func(content)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

def normalized_in_string_match(tocompare: str, comparedto: str)-> bool:
    """
    Remove all non-alphanumeric characters (hyphens, underscores, dots, slashes, colons)
    This turns both strings into pure blocks of letters and numbers
    """
    flat1 = re.sub(r'[^a-zA-Z0-9]', '', tocompare)
    flat2 = re.sub(r'[^a-zA-Z0-9]', '', comparedto)
    return flat1 in flat2



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api to work with templates """
from pathlib import Path
import hashlib
import shutil
import os
from cookiecutter.main import cookiecutter

from git_system_follower.logger import logger
from git_system_follower.errors import PackageApiError
from git_system_follower.package.package_info import get_gear_info
from git_system_follower.utils.tmpdir import tempdir, multi_tempdirs


__all__ = ['get_template_names', 'create_template', 'delete_template']


def get_template_names(script_dir: Path) -> tuple[str, ...]:
    """ Get available template names in package api """
    path = script_dir / 'templates'
    if not path.exists():
        raise PackageApiError(f'Template directory is missing. Path: {path}')

    return tuple(template.name for template in path.iterdir() if (path / template).is_dir())


@multi_tempdirs(2)
def create_template(
        script_dir: Path, template: str | None, target: Path, *,
        tmpdir: list[Path], variables: dict[str, str], is_force: bool,
        current_version_dir: Path | None = None
) -> None:
    logger.info(f'\t-> Creating project using {template} template')
    if template is None:
        logger.info('\t\tNo template specified. Skip operations')
        return
    path = script_dir / 'templates' / template

    if not path.exists():
        raise PackageApiError(f'Template is missing. Template path: {path}')

    new_version_path = tmpdir[0]
    current_version_path = tmpdir[1]
    cookiecutter(
        template=str(path), output_dir=new_version_path, no_input=True,
        extra_context=_get_extra_content(target, variables=variables)
    )
    if current_version_dir is not None:
        current_version_dir = current_version_dir / 'templates' / template
        cookiecutter(
            template=str(current_version_dir), output_dir=current_version_path, no_input=True,
            extra_context=_get_extra_content(target, variables=variables)
        )
    _copy_files(
        new_version_path / target.name, current_version_path / target.name, target,
        script_dir=script_dir, is_force=is_force
    )
    logger.info(f'\t\tSuccessful use template ({path})')


def _copy_files(source: Path, source_current: Path, target: Path, *, script_dir: Path, is_force: bool) -> None:
    paths = source.glob('**/*')
    for path in paths:
        if path.is_dir():
            continue
        relative_path = path.relative_to(source)
        target_path = target / relative_path
        current_path = source_current / relative_path
        if not target_path.exists():
            target_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(path, target_path)
            logger.info(f'\t\tFile {relative_path} does not exist. Copied it from template')
            continue

        path_hash = _calculate_hash(path)
        current_path_hash = _calculate_hash(current_path)
        target_path_hash = _calculate_hash(target_path)
        gear_info = get_gear_info(script_dir.parents[2], is_force=is_force)
        if path_hash == target_path_hash:
            logger.info(f'\t\tContent of {relative_path} file is same. Skip operations')
            continue
        elif path_hash != target_path_hash and current_path_hash == target_path_hash:
            logger.info(f'\t\tNo user changes for {relative_path} hence upgrading from template')
            shutil.copy(path, target_path)
            continue
        elif path_hash != target_path_hash and current_path_hash != target_path_hash:
            if gear_info['structure_type'] == 'simple' or gear_info['migrate']:
                shutil.copy(path, target_path)
            else:
                logger.warning(f'\t\tUser changes found for file {relative_path}. Cannot copy. Skip operations')
            continue

        if is_force:
            shutil.copy(path, target_path)
            logger.warning(f'\t\tFile {relative_path} already exists. Force flag enabled. '
                           f'Overwrote it from template')
            continue
        logger.warning(f'\t\tFile {relative_path} already exists. Cannot copy. Skip operations')


def _calculate_hash(path: Path) -> str:
    """ Calculating file content hash

    :param path: file path
    :return: file content hash
    """
    hash_func = hashlib.sha256()
    if Path(path).exists():
        with open(path, 'rb') as file:
            content = file.read().decode('utf-8-sig', errors='ignore')
            content = content.replace('\r\n', '\n').replace('\r', '\n')
            hash_func.update(content.encode('utf-8'))
            return hash_func.hexdigest()
    return "File not found"


@tempdir
def delete_template(
        script_dir: Path, template: str | None, target: Path, *,
        variables: dict[str, str], tmpdir: Path, is_force: bool
) -> None:
    logger.info(f'\t-> Deleting project files using {template} template')
    if template is None:
        logger.info('\t\tNo template specified. Skip operations')
        return
    path = script_dir / 'templates' / template
    if not path.exists():
        raise PackageApiError(f'Template is missing. Template path: {path}')

    cookiecutter(
        template=str(path), output_dir=tmpdir, no_input=True,
        extra_context=_get_extra_content(target, variables=variables)
    )
    _delete_files(tmpdir / target.name, target, is_force=is_force)
    logger.info(f'\t\tSuccessful delete files using template ({path})')


def _get_extra_content(target: Path, *, variables: dict[str, str]) -> dict[str, str]:
    content = {'gsf_repository_name': target.name}
    content.update(variables)
    return content


def _delete_files(source: Path, target: Path, *, is_force: bool) -> None:
    paths = source.glob('**/*')
    for path in paths:
        if path.is_dir():
            continue
        relative_path = path.relative_to(source)
        target_path = target / relative_path
        if not target_path.exists():
            logger.info(f'\t\tFile {relative_path} does not exist. Skip operations')
            continue

        path_hash, target_path_hash = _calculate_hash(path), _calculate_hash(target_path)
        if path_hash == target_path_hash:
            os.remove(target_path)
            logger.info(f'\t\tContent of {relative_path} file is same. Deleted this file')
            continue

        if is_force:
            os.remove(target_path)
            logger.warning(f'\t\tFile {relative_path} exists and does not match the template. '
                           f'Overwrote it from template')
            continue
        logger.warning(f'\t\tFile {relative_path} exists and does not match the template. '
                       f'Cannot delete. Skip operations')



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api with types for package developers  """
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from gitlab.v4.objects import Project
from git_system_follower.typings.cli import ExtraParam
from git_system_follower.package.cicd_variables import CICDVariable


__all__ = [
    'Parameters', 'SystemParameters', 'System',
    'CICDVariable', 'CICDVariables',
    'ExtraParam', 'ExtraParams'
]


@dataclass(frozen=True)
class SystemParameters:
    project: Project
    created_cicd_vars_names: list[str]
    script_dir: Path
    is_force: bool


@dataclass(frozen=True)
class System:
    host_domain: str


ExtraParamName = str
ExtraParams = dict[ExtraParamName, ExtraParam]

CICDVariableName = str
CICDVariables = dict[CICDVariableName, CICDVariable]


@dataclass(frozen=True)
class Parameters:
    __system_params: SystemParameters
    system: System
    workdir: Path
    extras: ExtraParams
    cicd_variables: CICDVariables
    all_cicd_variables: CICDVariables
    used_template: str
    template_variables: dict[str, str]
    current_version_dir: Optional[Path]
